# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import logging
from dataclasses import asdict, dataclass
from typing import Dict, List

from blue_krill.monitoring.probe.base import Issue, VirtualProbe

try:
    import pymysql
except ImportError as _e:
    raise ImportError('Error loading mysql module: %s.\n' 'Did you install pymysql?' % _e) from _e


logger = logging.getLogger(__name__)


@dataclass
class MySQLConfig:
    host: str
    port: int
    user: str
    password: str
    database: str


class MySQLProbe(VirtualProbe):
    """
    Usage:
        class SomeMySQLProbe(MySQLProbe):
            name: str = "some"
            config = MySQLConfig(...)
    """

    config: MySQLConfig

    def diagnose(self) -> List[Issue]:
        try:
            connection = pymysql.connect(**asdict(self.config))
        except pymysql.err.OperationalError:
            logger.exception("Can't connect to MySQL server on %s", self.config.host)
            return [Issue(fatal=True, description=f"Can't connect to MySQL server on {self.config.host}")]
        except Exception:  # pylint: disable=broad-except
            logger.exception("Unexpect exception occur when connecting to MySQL server on %s", self.config.host)
            return [Issue(fatal=True, description=f"Can't connect to MySQL server on {self.config.host}")]

        try:
            with connection.cursor() as cursor:
                sql = "SELECT 1"
                cursor.execute(sql)

            connection.commit()
        except Exception as e:  # pylint: disable=broad-except
            return [Issue(fatal=True, description=f"connection to mysql<{self.name}> failed: {e}")]
        finally:
            connection.close()
        return []


def transfer_django_db_settings(django_db_settings: Dict) -> MySQLConfig:
    """transfer django db settings to MySQLConfig"""

    return MySQLConfig(
        **{
            "host": django_db_settings['HOST'],
            "port": int(django_db_settings['PORT']),
            "user": django_db_settings['USER'],
            "password": django_db_settings['PASSWORD'],
            "database": django_db_settings['NAME'],
        }
    )



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
from os import PathLike
from typing import BinaryIO, ClassVar

from blue_krill.data_types.enum import EnumField, StructuredEnum


class SignatureType(str, StructuredEnum):
    DOWNLOAD = EnumField("DOWNLOAD", label="下载")
    UPLOAD = EnumField("UPLOAD", label="上传")


class BlobStore:
    """BlobStore Protocol, should be implemented by subclass."""

    STORE_TYPE: ClassVar[str]

    def __init__(self, bucket: str):
        self.bucket = bucket

    def get_client(self):
        raise NotImplementedError

    def upload_file(self, filepath: PathLike, key: str, allow_overwrite: bool = True, **kwargs):
        """Upload package to server

        :param PathLike filepath: The path to the file to upload.
        :param str key: key to store the package.
        :param bool allow_overwrite: whether to overwrite the original file
        :raise BlobAlreadyExists: if the key already exists.
        :raise UploadFailedError: if the server return error when uploading
        """
        raise NotImplementedError

    def upload_fileobj(self, fh: BinaryIO, key: str, allow_overwrite: bool = True, **kwargs):
        """Upload package to server

        :param BinaryIO fh: A file-like object to upload. At a minimum, it must
                            implement the `read` method, and must return bytes.
        :param str key: The name of the key to upload to.
        :param bool allow_overwrite: whether to overwrite the original file
        :raise BlobAlreadyExists: if the key already exists.
        :raise UploadFailedError: if the server return error when uploading
        """
        raise NotImplementedError

    def download_file(self, key: str, filepath: PathLike, *args, **kwargs) -> PathLike:
        """Download file to filepath
        :param str key: The name of the key to download from.
        :param PathLike filepath: The path to the file to download to.
        :return path to download.

        :raise DownloadFailedError: If due to server error or client error (such as disk full)
        """
        raise NotImplementedError

    def download_fileobj(self, key: str, fh: BinaryIO, *args, **kwargs):
        """Download file to fh
        :param str key: The name of the key to download from.
        :param BinaryIO fh: The fileobj to the file to download to.

        :raise DownloadFailedError: If due to server error or client error (such as disk full)
        """
        raise NotImplementedError

    def delete_file(self, key: str, *args, **kwargs):
        """Delete file in filepath

        :param str key: The name of the key to delete.
        """
        raise NotImplementedError

    def get_file_metadata(self, key: str):
        """Get file's metadata
        :param str key: The name of the key to read.
        """
        raise NotImplementedError

    def generate_presigned_url(
        self, key: str, expires_in: int, signature_type: SignatureType = SignatureType.DOWNLOAD, *args, **kwargs
    ) -> str:
        """Generate pre-signed url to share blob in store.

        :param str key: key storing the package to be downloaded.
        :param int expires_in: when will the url expire.
        :param SignatureType signature_type: the permision for the presigned url.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import json
import logging
import time

from blue_krill.encoding import force_text

logger = logging.getLogger(__name__)


class StreamChannel(object):
    """A simple stream channel implemention"""

    default_expires_seconds = 3600 * 24

    def __init__(self, channel_id, redis_db=None, expires_seconds=None):
        self.channel_id = channel_id
        self.keys = KeyManager(channel_id)
        self.redis_db = redis_db
        self.expires_seconds = expires_seconds or self.default_expires_seconds

    def initialize(self):
        """Initialize channel"""
        if self.redis_db.exists(self.keys.counter):
            return

        # Always update state first
        self.redis_db.set(self.keys.state, 'open')
        self.publish('init')

        # Set expires
        pipe = self.redis_db.pipeline()
        for key in self.keys.entities():
            pipe.expire(key, self.expires_seconds)
        pipe.execute()

    def publish_msg(self, message):
        self.publish('msg', data=message)

    def publish(self, event, data=''):
        index = self.redis_db.incr(self.keys.counter)
        payload = {'id': index, 'event': event, 'data': data}
        self.redis_db.rpush(self.keys.history, json.dumps(payload))
        self.redis_db.publish(self.keys.channel, json.dumps(payload))

    def close(self):
        self.redis_db.set(self.keys.state, 'closed', self.expires_seconds)
        self.publish('close')

    def destroy(self):
        """Destory this channel, every history events will be deleted!"""
        pipe = self.redis_db.pipeline()
        for key in self.keys.entities():
            pipe.delete(key)
        pipe.execute()

    def __str__(self):
        return 'StreamChannel: {}'.format(self.channel_id)


class StreamChannelSubscriber(object):
    """Subscriber for StreamChannel"""

    def __init__(self, channel_id, redis_db=None):
        self.channel_id = channel_id
        self.keys = KeyManager(channel_id)
        self.redis_db = redis_db

        self.sub_pipe = self.redis_db.pubsub(ignore_subscribe_messages=True)
        self.sub_pipe.subscribe(self.keys.channel)

        self.update_channel_state()

    def get_channel_state(self):
        return self._channel_state

    def update_channel_state(self):
        self._channel_state = self.read_channel_state()

    def read_channel_state(self):
        state = self.redis_db.get(self.keys.state)
        state = force_text(state)
        if state is None:
            return 'none'
        elif state == 'open':
            return 'open'
        elif state == 'closed':
            return 'closed'
        return 'unknown'

    def is_closed(self):
        return self.get_channel_state() == 'closed'

    def get_history_events(self, last_event_id=0, ignore_special=True):
        """Get history events

        :param int last_event_id: If given, result will start from last_event_id
        """
        results = []
        for item in self.redis_db.lrange(self.keys.history, last_event_id, -1):
            event = json.loads(item)
            # Update Channel status if event is special
            if self.is_special_event(event):
                self.update_channel_state()
            if ignore_special and self.is_special_event(event):
                continue
            results.append(event)
        return results

    def get_events(self, last_event_id=0, wait=0.05, ignore_special=True):
        """Get all history events and follow new one.

        :param int last_event_id: Ignore every events whose id is lower than this
        :param int wait: Wait for seconds for every cycle
        """
        # TODO: If a channel is not open for a long time, raise an exception
        events = self.get_history_events(last_event_id=last_event_id, ignore_special=ignore_special)
        for event in events:
            yield event

        max_event_id = events[-1]['id'] if events else last_event_id
        while True:
            if self.is_closed():
                break
            event = self.get_event(block=False)
            # Be nice to your system, sleep for a while
            if not event:
                time.sleep(wait)
                continue
            # Ignore event that already fetched from history events
            if event['id'] <= max_event_id:
                continue
            if ignore_special and self.is_special_event(event):
                continue
            yield event

    def get_event(self, block=False):
        """Get event from subscribe

        :param block bool: if True, will use .listen method to do a sync read
        """
        if block:
            _data = next(self.sub_pipe.listen())
        else:
            _data = self.sub_pipe.get_message()
        if not _data:
            return

        data = json.loads(_data['data'])

        # Update Channel status if event is special
        if self.is_special_event(data):
            self.update_channel_state()
        return data

    def is_special_event(self, event):
        return event['event'] in ('init', 'close')

    def close(self):
        self.sub_pipe.close()

    def __str__(self):
        return 'StreamChannelSubscriber: {}'.format(self.channel_id)


class KeyManager(object):
    """Redis key manager for channel"""

    default_key_prefix = 'p3::'

    def __init__(self, channel_id, prefix=None):
        prefix = prefix or self.default_key_prefix
        self.state = '{}evtch::{}::sta'.format(prefix, channel_id)
        self.counter = '{}evtch::{}::cnt'.format(prefix, channel_id)
        self.history = '{}evtch::{}::his'.format(prefix, channel_id)
        self.channel = '{}evtch::{}::cha'.format(prefix, channel_id)

    def entities(self):
        return [self.state, self.counter, self.history]



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
from cryptography.fernet import Fernet

from blue_krill.encoding import force_bytes, force_text
from blue_krill.encrypt.utils import get_default_secret_key


class EncryptHandler:
    def __init__(self, secret_key=get_default_secret_key()):
        self.secret_key = secret_key
        self.f = Fernet(self.secret_key)

    def encrypt(self, text: str) -> str:
        if self.Header.contain_header(text):
            return text

        b_text = force_bytes(text)
        return self.Header.add_header(force_text(self.f.encrypt(b_text)))

    def decrypt(self, encrypted: str) -> str:
        encrypted = self.Header.strip_header(encrypted)

        b_encrypted = force_bytes(encrypted)
        return force_text(self.f.decrypt(b_encrypted))

    class Header:
        HEADER = "bkcrypt$"

        @classmethod
        def add_header(cls, text: str):
            return cls.HEADER + text

        @classmethod
        def strip_header(cls, text: str):
            # 兼容无 header 加密串
            if not cls.contain_header(text):
                return text

            return text[len(cls.HEADER) :]

        @classmethod
        def contain_header(cls, text: str) -> bool:
            return text.startswith(cls.HEADER)



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import logging
import socket
from typing import List, NamedTuple

from blue_krill.monitoring.probe.base import Issue, VirtualProbe

logger = logging.getLogger(__name__)


class InternetAddress(NamedTuple):
    host: str
    port: int


class TCPProbe(VirtualProbe):
    """
    Usage:
        class SomeTCPProbe(TCPProbe):
            name: str = "some"
            address: InternetAddress = InternetAddress(host="localhost", port=8080)

    """

    timeout: int = 30
    address: InternetAddress

    def diagnose(self) -> List[Issue]:
        try:
            sock = socket.create_connection(address=self.address, timeout=self.timeout)
            sock.close()
        except socket.timeout as e:
            logger.exception("Connect %s Timeout", self.address)
            return [Issue(fatal=True, description=f"Connect {self.address} Timeout, detail: {str(e)}")]
        except Exception as e:  # pylint: disable=broad-except
            logger.exception("Unknown Exception when connecting to %s", self.address)
            return [Issue(fatal=True, description=f"Unknown Exception<{type(e).__name__}>, detail: {str(e)}")]
        return []



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""

import re
from functools import partial
from typing import Callable, Iterable, Optional

__all__ = ["RegexCubingHelper", "CommonCaseRegexPatterns", "CommonCaseConvertor", "shortcuts"]


class RegexCubingHelper:
    """The best way to resolve a Rubik's cube is to take it apart and put it back together."""

    def __init__(self, patterns: Iterable[str], preprocessor: Optional[Callable[[str], str]] = None):
        self._pattern = re.compile("|".join(patterns))
        self._preprocessor = preprocessor

    def cubing(self, string: str, transform_fn: Callable[[Iterable[str]], Iterable[str]], sep: str) -> str:
        """Join the string with the given separator and transform the parts."""

        if self._preprocessor:
            string = self._preprocessor(string)

        parts = (i for i in self._pattern.split(string) if i)
        return sep.join((i for i in transform_fn(parts) if i))

    def cubing_capitalize_case(self, string: str, sep: str) -> str:
        """Join the string with the given separator and capitalize the parts."""

        return self.cubing(string, lambda parts: (i.capitalize() for i in parts), sep)

    def cubing_lower_case(self, string: str, sep: str) -> str:
        """Join the string with the given separator and lower case the parts."""

        return self.cubing(string, lambda parts: (i.lower() for i in parts), sep)

    def cubing_upper_case(self, string: str, sep: str) -> str:
        """Join the string with the given separator and upper case the parts."""

        return self.cubing(string, lambda parts: (i.upper() for i in parts), sep)


class CommonCaseRegexPatterns:
    """The regex pattern for common case."""

    CAMELCASE = r"(?<=[^A-Z])(?=[A-Z])"
    SNAKECASE = r"_+"
    DASHCASE = r"-+"
    DOTCASE = r"\.+"
    SPACECASE = r" +"


class CommonCaseConvertor:
    """The convertor for common case."""

    def __init__(self, patterns: Iterable[str], preprocessor: Optional[Callable[[str], str]] = None):
        self._helper = RegexCubingHelper(patterns, preprocessor)

    def to_camel_case(self, string: str) -> str:
        """
        Convert the string to camel case, like this:
        >>> convertor = CommonCaseConvertor([CommonCaseRegexPatterns.SPACECASE])
        >>> convertor.to_camel_case("Cubing Case")
        'CubingCase'
        """

        return self._helper.cubing_capitalize_case(string, "")

    def to_lower_camel_case(self, string: str) -> str:
        """Convert the string to lower camel case, like this:
        >>> convertor = CommonCaseConvertor([CommonCaseRegexPatterns.SPACECASE])
        >>> convertor.to_lower_camel_case("Cubing Case")
        'cubingCase'
        """

        def transform(parts):
            for index, part in enumerate(parts):
                if index == 0:
                    yield part.lower()
                else:
                    yield part.capitalize()

        return self._helper.cubing(string, transform, "")

    def to_lower_snake_case(self, string: str) -> str:
        """
        Convert the string to lower snake case, like this:
        >>> convertor = CommonCaseConvertor([CommonCaseRegexPatterns.SPACECASE])
        >>> convertor.to_lower_snake_case("Cubing Case")
        'cubing_case'
        """

        return self._helper.cubing_lower_case(string, "_")

    def to_upper_snake_case(self, string: str) -> str:
        """
        Convert the string to upper snake case, like this:
        >>> convertor = CommonCaseConvertor([CommonCaseRegexPatterns.SPACECASE])
        >>> convertor.to_upper_snake_case("Cubing Case")
        'CUBING_CASE'
        """

        return self._helper.cubing_upper_case(string, "_")

    def to_lower_dash_case(self, string: str) -> str:
        """
        Convert the string to lower dash case, like this:
        >>> convertor = CommonCaseConvertor([CommonCaseRegexPatterns.SPACECASE])
        >>> convertor.to_lower_dash_case("Cubing Case")
        'cubing-case'
        """

        return self._helper.cubing_lower_case(string, "-")

    def to_upper_dash_case(self, string: str) -> str:
        """Convert the string to upper dash case, like this:
        >>> convertor = CommonCaseConvertor([CommonCaseRegexPatterns.SPACECASE])
        >>> convertor.to_upper_dash_case("Cubing Case")
        'CUBING-CASE'
        """

        return self._helper.cubing_upper_case(string, "-")

    def to_lower_dot_case(self, string: str) -> str:
        """Convert the string to lower dot case, like this:
        >>> convertor = CommonCaseConvertor([CommonCaseRegexPatterns.SPACECASE])
        >>> convertor.to_lower_dot_case("Cubing Case")
        'cubing.case'
        """

        return self._helper.cubing_lower_case(string, ".")

    def to_capitalize_dot_case(self, string: str) -> str:
        """Convert the string to upper dot case, like this:
        >>> convertor = CommonCaseConvertor([CommonCaseRegexPatterns.SPACECASE])
        >>> convertor.to_capitalize_dot_case("Cubing Case")
        'Cubing.Case'
        """

        return self._helper.cubing_capitalize_case(string, ".")

    def to_lower_space_case(self, string: str) -> str:
        """Convert the string to lower space case, like this:
        >>> convertor = CommonCaseConvertor([CommonCaseRegexPatterns.SPACECASE])
        >>> convertor.to_lower_space_case("Cubing Case")
        'cubing case'
        """

        return self._helper.cubing_lower_case(string, " ")


_shortcuts_patterns = [
    CommonCaseRegexPatterns.SNAKECASE,
    CommonCaseRegexPatterns.DASHCASE,
    CommonCaseRegexPatterns.DOTCASE,
    CommonCaseRegexPatterns.SPACECASE,
]
_shortcuts_preprocessor = None

try:
    # this pattern must be a zero-length assertion, test if it contains the underlying bug
    re.split("(?=.)", "")
except ValueError:
    # convert the string from camel case to space case for python 3.6, details: https://bugs.python.org/issue43222
    _shortcuts_preprocessor = partial(re.compile(f"({CommonCaseRegexPatterns.CAMELCASE})").sub, " ")
else:
    # for python 3.7+, it's ok to handle camel case directly
    _shortcuts_patterns.append(CommonCaseRegexPatterns.CAMELCASE)


shortcuts = CommonCaseConvertor(_shortcuts_patterns, _shortcuts_preprocessor)



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import os

MULTIPROC_ENV_NAME = 'prometheus_multiproc_dir'
DEFAULT_DATA_PATH = '_prometheus_data'


def setup_multiproc_dir(data_path: str = DEFAULT_DATA_PATH):
    """Setup multi processes directory

    :param data_path: if MULTIPROC_ENV_NAME has already been configured in env vars, then the
        data_path parameter won't be used at all
    """
    dirname = os.environ.setdefault(MULTIPROC_ENV_NAME, data_path)
    try:
        os.mkdir(dirname)
    except FileExistsError:
        pass
    except IOError as e:
        print(f'unable to make prometheus data directory: {e}')



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import os
from typing import Optional

from cryptography.fernet import InvalidToken
from environ import Env

from blue_krill.encrypt.utils import decrypt_string


class EncryptedEnviron:
    """A wrapper class which reads encrypted values from os.environ

    :param key: if not given, decryptor may use default secret key from settings
    :param decryptor: a callable object that decrypt the value by key
    """

    def __init__(self, environ=None, key: Optional[bytes] = None, decryptor=decrypt_string):
        self.environ = environ or os.environ
        self.key = key
        self.decryptor = decryptor

    def __getitem__(self, key) -> str:
        """Overwrite the default env[var_name] behavior with an extra decryption step

        :raises: ValueError when value was not encrypted by fernet
        """
        _value = self.environ[key]

        try:
            return self.decryptor(_value, key=self.key)
        except InvalidToken:
            raise ValueError(f'Var "{key}" was not encrypted or token is invalid, please use plain text mode instead')


class SecureEnv(Env):
    """Allow to read value from encrypted environment variables"""

    ENVIRON_CLS = EncryptedEnviron

    def set_secure_key(self, key: bytes):
        """Set a different key for decryption"""
        self.ENVIRON = self.ENVIRON_CLS(key=key)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.ENVIRON = self.ENVIRON_CLS()



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import logging
import os
import secrets
from collections import namedtuple
from typing import Callable, Dict, Iterable

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.handlers.wsgi import WSGIHandler, get_path_info
from prometheus_client import CONTENT_TYPE_LATEST, CollectorRegistry, generate_latest, make_wsgi_app, multiprocess

from .setups import MULTIPROC_ENV_NAME

logger = logging.getLogger(__name__)

# Validate settings
try:
    METRIC_CLIENT_TOKEN_DICT = settings.METRIC_CLIENT_TOKEN_DICT
    if not isinstance(METRIC_CLIENT_TOKEN_DICT, dict):
        raise ValueError('client token dict must be dict')
except (AttributeError, ValueError) as e:
    raise ImproperlyConfigured("METRIC_CLIENT_TOKEN_DICT is not properly configured") from e


class WSGIResponse(namedtuple('WSGIResponse', 'status data')):
    """A simple wsgi response wrapper class"""

    def write_to(self, start_response: Callable) -> Iterable:
        response_headers = [('Content-type', CONTENT_TYPE_LATEST), ('Content-Length', str(len(self.data)))]
        start_response(self.status, response_headers)
        return iter([self.data])


class PrometheusExposeHandler(WSGIHandler):
    """WSGI middleware that intercepts calls from the prometheus"""

    def __init__(self, application, exposed_path="/metrics", extra_registries=None):
        """A prometheus wsgi handler for django

        :param application: wsgi application
        :param exposed_path: path of metrics
        :param extra_registries: Extra registry objects to collect metrics from, default to None
        """
        self.application = application
        self.exposed_path = exposed_path
        extra_registries = extra_registries or []
        if os.environ.get(MULTIPROC_ENV_NAME):
            logger.info("enable prometheus using multi processes mode")

            def app(environ, start_response):
                registry = CollectorRegistry()
                multiprocess.MultiProcessCollector(registry)

                data_chunks = []

                # Combine response from 2 sources:
                #   1. the empty 'registry' object, which will load data from all files in multiproc directory
                #   2. all given extra registry objects, usually registed with custom collectors
                for r in [registry, *extra_registries]:
                    data_chunks.append(generate_latest(r))
                data = b''.join(data_chunks)

                return WSGIResponse('200 OK', data).write_to(start_response)

        else:
            logger.info("enable prometheus using single process mode")
            app = make_wsgi_app()

        self.prometheus_wsgi_app = app
        self.authenticater = HandlerAuthenticater(METRIC_CLIENT_TOKEN_DICT)
        super().__init__()

    def _should_handle(self, path):
        """Checks if the path should be handled."""
        return path.startswith(self.exposed_path)

    def __call__(self, environ, start_response):
        # Hand non-metrics requests to Django
        if not self._should_handle(get_path_info(environ)):
            return self.application(environ, start_response)

        request = self.request_class(environ)
        if self.authenticater.validate_request(request):
            return self.prometheus_wsgi_app(environ, start_response)

        return self.authenticater.generate_failed_response().write_to(start_response)


class HandlerAuthenticater:
    """Authenticater for prometheus handler"""

    key_username, key_password = 'client_id', 'user_token'

    def __init__(self, credentials: Dict[str, str]):
        self.credentials = credentials

    def validate_request(self, request) -> bool:
        """Validate a given request"""
        username = request.GET.get(self.key_username, '')
        password = request.GET.get(self.key_password, '')
        if not (username and password):
            return False
        return secrets.compare_digest(password, self.credentials.get(username, ''))

    def generate_failed_response(self) -> WSGIResponse:
        """Generate a failed response"""
        data = b'permission denied, no valid client_id/token provided'
        return WSGIResponse('403 FORBIDDEN', data)



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import logging
from binascii import Error as PaddingError

from jwt.utils import base64url_decode

logger = logging.getLogger(__name__)


def validate_jwt_token(s: str) -> bool:
    """
    Test if a given string might be a JWT token, more on JWT: https://jwt.io/introduction
    """
    parts = s.split('.')
    if len(parts) != 3:
        return False

    try:
        # JWT: {header}.{payload}.{signature}
        # The "header" and "payload" parts were always b64encoded
        base64url_decode(parts[0])
        base64url_decode(parts[1])
    except (PaddingError, ValueError):
        return False
    return True


def get_paas_service_jwt_clients():
    try:
        from django.conf import settings

        return settings.PAAS_SERVICE_JWT_CLIENTS
    except ImportError:
        logger.exception("you should supply paas service jwt clients")
        raise



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import logging
from os import PathLike, getenv
from typing import Any, BinaryIO, List
from urllib.parse import urljoin

import curlify
import requests
from requests.adapters import HTTPAdapter
from requests.auth import HTTPBasicAuth

from blue_krill.data_types.enum import EnumField, StructuredEnum
from blue_krill.storages.blobstore.base import BlobStore, SignatureType
from blue_krill.storages.blobstore.exceptions import (
    DownloadFailedError,
    ObjectAlreadyExists,
    RequestError,
    UploadFailedError,
)

MAX_RETRIES = 2
logger = logging.getLogger(__name__)

TIMEOUT_THRESHOLD = float(getenv("BKREPO_TIMEOUT_THRESHOLD", 30))


def _validate_resp(response: requests.Response) -> Any:
    """校验响应体"""
    try:
        logger.info("Calling BkRepo, the equivalent curl command: %s", curlify.to_curl(response.request))
    except Exception:  # pylint: disable=broad-except
        pass

    try:
        data = response.json()
    except Exception as e:  # pylint: disable=broad-except
        raise RequestError(str(e), code="Unknown", response=response) from e

    if data.get("code") != 0:
        raise RequestError(data.get("message"), code=data.get("code"), response=response)
    return data.get("data", {})


class RepositoryType(str, StructuredEnum):
    GENERIC = EnumField("GENERIC", label="通用二进制文件仓库")
    DOCKER = EnumField("DOCKER", label="Docker仓库")
    MAVEN = EnumField("MAVEN", label="Maven仓库")
    PYPI = EnumField("PYPI", label="Pypi仓库")
    NPM = EnumField("NPM", label="Npm仓库")
    HELM = EnumField("HELM", label="Helm仓库")
    COMPOSER = EnumField("COMPOSER", label="COMPOSER仓库")
    RPM = EnumField("RPM", label="Rpm仓库")


class BKRepoManager:
    """蓝鲸 bkrepo 管理端"""

    def __init__(
        self,
        endpoint_url: str,
        username: str,
        password: str,
        **kwargs,
    ):
        # endpoint can not endswith '/'
        self.endpoint_url = endpoint_url.rstrip("/")
        self.username = username
        self.password = password
        self._max_retries = kwargs.get("max_retries", MAX_RETRIES)

    def get_client(self) -> requests.Session:
        session = requests.session()
        session.auth = HTTPBasicAuth(username=self.username, password=self.password)
        session.mount("http://", HTTPAdapter(max_retries=self._max_retries))
        session.mount("https://", HTTPAdapter(max_retries=self._max_retries))
        return session

    def create_user_to_repo(
        self, username: str, password: str, association_users: List[str], project: str, repo: str
    ) -> bool:
        """创建用户到仓库管理员

        :params username str: 用户名
        :params password str: 密码
        :params association_users List[str]: 关联的真实用户
        :params repo str: 关联的仓库名称
        """
        client = self.get_client()
        url = urljoin(self.endpoint_url, '/auth/api/user/create/repo')
        data = {
            "admin": True,
            "name": username,
            "pwd": password,
            "userId": username,
            "asstUsers": association_users,
            "group": False,
            "projectId": project,
            "repoName": repo,
        }
        return _validate_resp(client.post(url, json=data, timeout=TIMEOUT_THRESHOLD))

    def update_user(self, username: str, password: str, association_users: List[str]):
        """更新用户信息"""
        client = self.get_client()
        url = urljoin(self.endpoint_url, f'/auth/api/user/{username}')
        data = {"admin": True, "name": username, "pwd": password, "asstUsers": association_users}
        return _validate_resp(client.put(url, json=data, timeout=TIMEOUT_THRESHOLD))

    def delete_user(self, username: str):
        """删除用户"""
        client = self.get_client()
        url = urljoin(self.endpoint_url, f'/auth/api/user/{username}')
        return _validate_resp(client.delete(url, timeout=TIMEOUT_THRESHOLD))

    def create_repo(self, project: str, repo: str, repo_type: str = RepositoryType.GENERIC, public: bool = False):
        """创建仓库

        :param public bool: 是否公开读, 当 public 为 True 时, 代表公开读私有写; 当 public 为 False 时, 代表私有读写
        """
        client = self.get_client()
        url = urljoin(self.endpoint_url, '/repository/api/repo/create')
        data = {
            "projectId": project,
            "name": repo,
            "type": RepositoryType(repo_type).value,
            "category": "LOCAL",
            "public": public,
            "description": "no description",
            "configuration": None,
            "storageCredentialsKey": None,
        }
        return _validate_resp(client.post(url, json=data, timeout=TIMEOUT_THRESHOLD))

    def delete_repo(self, project: str, repo: str, forced: bool = False):
        """删除仓库

        :params repo str: 仓库名
        :params forced bool: 是否强制删除, 如果为false，当仓库中存在文件时，将无法删除仓库
        """
        client = self.get_client()
        url = urljoin(self.endpoint_url, f'/repository/api/repo/delete/{project}/{repo}?forced={forced}')
        return _validate_resp(client.delete(url, timeout=TIMEOUT_THRESHOLD))

    # 以下是项目无关的管理接口

    def create_project(self, project: str):
        """创建项目

        :params project str: 项目名
        """
        client = self.get_client()
        url = urljoin(self.endpoint_url, "/repository/api/project/create")
        data = {"name": project, "displayName": project, "description": ""}
        return _validate_resp(client.post(url, json=data, timeout=TIMEOUT_THRESHOLD))

    def create_user_to_project(self, username: str, password: str, association_users: List[str], project: str) -> bool:
        """创建用户到项目管理员

        :params username str: 用户名
        :params password str: 密码
        :params association_users List[str]: 关联的真实用户
        :params project str: 关联的项目名称
        """
        client = self.get_client()
        url = urljoin(self.endpoint_url, '/auth/api/user/create/project')
        data = {
            "admin": True,
            "name": username,
            "pwd": password,
            "userId": username,
            "asstUsers": association_users,
            "group": False,
            "projectId": project,
        }
        return _validate_resp(client.post(url, json=data, timeout=TIMEOUT_THRESHOLD))


class BKGenericRepo(BlobStore):
    """蓝鲸通用二进制仓库."""

    STORE_TYPE = "bkrepo"

    def __init__(
        self,
        bucket: str,
        project: str,
        endpoint_url: str,
        username: str,
        password: str,
        **kwargs,
    ):
        super().__init__(bucket)
        self.project = project
        # endpoint can not endswith '/'
        self.endpoint_url = endpoint_url.rstrip("/")
        self.username = username
        self.password = password
        self._max_retries = kwargs.get("max_retries", MAX_RETRIES)

    def get_client(self) -> requests.Session:
        session = requests.session()
        session.auth = HTTPBasicAuth(username=self.username, password=self.password)
        session.mount("http://", HTTPAdapter(max_retries=self._max_retries))
        session.mount("https://", HTTPAdapter(max_retries=self._max_retries))
        return session

    def upload_file(self, filepath: PathLike, key: str, allow_overwrite: bool = True, **kwargs):
        """上传通用制品文件

        :param PathLike filepath: 需要上传文件的路径
        :param str key: 文件完整路径
        :param bool allow_overwrite: 是否覆盖已存在文件
        """
        with open(filepath, "rb") as fh:
            self.upload_fileobj(fh, key=key, allow_overwrite=allow_overwrite, timeout=TIMEOUT_THRESHOLD, **kwargs)

    def upload_fileobj(self, fh: BinaryIO, key: str, allow_overwrite: bool = True, **kwargs):
        """上传通用制品文件

        :param BinaryIO fh: 文件句柄
        :param str key: 文件完整路径
        :param bool allow_overwrite: 是否覆盖已存在文件
        """
        client = self.get_client()
        url = urljoin(self.endpoint_url, f'/generic/{self.project}/{self.bucket}/{key}')
        src = getattr(fh, "name", "<memory>")
        headers = {"X-BKREPO-OVERWRITE": str(allow_overwrite)}

        try:
            resp = client.put(url, headers=headers, data=fh, timeout=TIMEOUT_THRESHOLD)
            _validate_resp(resp)
        except RequestError as e:
            # 250107: 请求资源已经存在
            # 251012: Node Existed
            if str(e.code) in ["250107", "251012"]:
                raise ObjectAlreadyExists(e.message, e.code, e.response) from e
            logger.exception("Request success, but the server rejects the upload request.")
            raise UploadFailedError(key=key, src=src) from e
        except Exception as e:
            logger.exception("An unexpected exception occurred, detail: %s", e)
            raise UploadFailedError(key=key, src=src) from e

    def download_file(self, key: str, filepath: PathLike, *args, **kwargs) -> PathLike:
        """下载通用制品文件

        :param str key: 文件完整路径
        :param PathLike filepath: 文件下载的路径
        """
        with open(filepath, mode="wb") as fh:
            self.download_fileobj(key, fh)
        return filepath

    def download_fileobj(self, key: str, fh, *args, **kwargs):
        """下载通用制品文件

        :param str key: 文件完整路径
        :param IO fh: 文件句柄
        """
        client = self.get_client()
        url = urljoin(self.endpoint_url, f'/generic/{self.project}/{self.bucket}/{key}')
        dest = getattr(fh, "name", "<memory>")
        try:
            resp = client.get(url, stream=True, timeout=TIMEOUT_THRESHOLD)
            logger.info("Calling BkRepo, the equivalent curl command: %s", curlify.to_curl(resp.request))
        except Exception as e:
            logger.exception("Fail to init request to BkRepo when calling '%s'", url)
            raise DownloadFailedError(key=key, dest=dest) from e

        if not resp.ok:
            logger.exception(
                "Request success, but the server rejects the download request, please check by those curl command: %s",
                curlify.to_curl(resp.request),
            )
            raise DownloadFailedError(key=key, dest=dest) from RequestError(
                str("下载制品文件失败"), code=str(resp.status_code), response=resp
            )

        try:
            for chunk in resp.iter_content(chunk_size=512):
                if chunk:
                    fh.write(chunk)
        except Exception as e:
            logger.exception("File save failed, detail %s", e)
            raise DownloadFailedError(key=key, dest=dest) from e

    def delete_file(self, key: str, *args, **kwargs):
        """删除通用制品文件

        :param str key: 文件完整路径
        """
        client = self.get_client()
        url = urljoin(self.endpoint_url, f'/generic/{self.project}/{self.bucket}/{key}')
        resp = client.delete(url, timeout=TIMEOUT_THRESHOLD)
        return _validate_resp(resp)

    def get_file_metadata(self, key, *args, **kwargs):
        """获取通用制品文件头部信息

        :param str key: 文件完整路径
        """
        client = self.get_client()
        url = urljoin(self.endpoint_url, f'/generic/{self.project}/{self.bucket}/{key}')
        resp = client.head(url, timeout=TIMEOUT_THRESHOLD)
        if resp.status_code == 200:
            return resp.headers
        raise RequestError("Can't get file head info", code=str(resp.status_code), response=resp)

    def generate_presigned_url(
        self, key: str, expires_in: int, signature_type: SignatureType = SignatureType.DOWNLOAD, *args, **kwargs
    ) -> str:
        """创建临时访问url

        :param str key: 授权路径
        :param int expires_in: token 有效时间，单位秒，小于等于 0 则永久有效
        :param str signature_type: 签名类型。UPLOAD:允许上传, DOWNLOAD: 允许下载。
        :param str token_type: [deprecated] token类型。UPLOAD:允许上传, DOWNLOAD: 允许下载, ALL: 同时允许上传和下载。
        """
        client = self.get_client()
        url = urljoin(self.endpoint_url, '/generic/temporary/url/create')

        token_type = signature_type.value
        if "token_type" in kwargs:
            logger.warning("[token_type] is deprecated. Please use signature_type instead.")
            token_type = kwargs["token_type"]

        resp = client.post(
            url,
            json={
                'projectId': self.project,
                'repoName': self.bucket,
                'fullPathSet': [key],
                'expireSeconds': expires_in,
                'type': token_type,
            },
            timeout=TIMEOUT_THRESHOLD,
        )
        data = _validate_resp(resp)
        return data[0]["url"]



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import logging
from typing import Dict, List

from blue_krill.monitoring.probe.base import Issue, VirtualProbe

try:
    import requests
except ImportError as _e:
    raise ImportError('Error loading http module: %s.\n' 'Did you install requests?' % _e) from _e


logger = logging.getLogger(__name__)


class HttpProbe(VirtualProbe):
    """
    A universal http probe, which send a GET request to url and only validate status code from response.

    Usage:
        class SomeHttpProbe(HttpProbe):
            name: str = "some"
            url: str = "http://localhost/ping"

        class SomeHttpsWithoutCa(HttpProbe):
            name: str = "some"
            url: str = "http://localhost/ping"
            verify: bool = False

        class SomeHttpWithAuth(HttpProbe):
            name: str = "some"
            url: str = "http://localhost/ping"
            params: Dict = {"token": "dummy"}
            headers: Dict = {"Authorization": "Basic YWxhZGRpbjpvcGVuc2VzYW1l"}

    """

    url: str
    headers: Dict = {}
    params: Dict = {}
    verify: bool = True
    timeout: int = 30
    desired_code: int = requests.codes.ok

    def diagnose(self) -> List[Issue]:
        try:
            response = requests.get(
                url=self.url, params=self.params, headers=self.headers, timeout=self.timeout, verify=self.verify
            )
        except requests.Timeout as e:
            logger.exception("Request %s timeout", self.url)
            return [Issue(fatal=True, description=f"Request {self.url} timeout, detail: {e}")]
        except requests.RequestException as e:
            logger.exception("An exception occurs when requesting %s", self.url)
            return [Issue(fatal=True, description=str(e))]
        return self.validate_response(response)

    def validate_response(self, response: requests.Response) -> List[Issue]:
        if response.status_code != self.desired_code:
            return [Issue(fatal=True, description=f"Undesired status code<{response.status_code}>")]
        return []


class BKHttpProbe(HttpProbe):
    """A specialized http probe for internal projects, which validate a formatted `healthz` response.

    Example Formatted Response:
        {result: bool, message: str}

    When bool('result') is False, report an issue with 'message'.
    """

    def validate_response(self, response: requests.Response) -> List[Issue]:
        if response.status_code != self.desired_code:
            return [Issue(fatal=True, description=f"Undesired status code<{response.status_code}>")]

        try:
            data = response.json()
        except Exception:  # pylint: disable=broad-except
            logger.exception("invalid data: %s", response.content)
            return [Issue(fatal=True, description=f"invalid data: {response.content!r}")]

        if not bool(data.get("result")):
            return [Issue(fatal=True, description=data.get("message", "missing message."))]
        return []



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import six
import logging
import time
from dataclasses import dataclass
from typing import Dict, Optional

import jwt
from requests.auth import AuthBase

logger = logging.getLogger(__name__)

DEFAULT_ALGORITHM = 'HS256'

DEFAULT_ROLE = 'default'


@dataclass
class JWTAuthConf:
    """JWT conf object

    :param extra_payload: extra key value pairs in generated payload
    """

    iss: str
    key: str
    algorithm: str = DEFAULT_ALGORITHM
    role: str = DEFAULT_ROLE


class ClientJWTAuth(AuthBase):
    """Auto sign and inject a JWT token in the request headers

    :param auth_conf: JWT Auth instance
    :param prefix: Prefix string of header, default to "Bearer"
    :param expires_in: After how many seconds, token will be considered expired, default to 3600
    """

    _default_expires_in = 3600

    def __init__(self, auth_conf: JWTAuthConf, prefix: str = 'Bearer', expires_in: int = None):
        self.auth_conf = auth_conf
        self.prefix = prefix
        self.expires_in = expires_in or self._default_expires_in

    def __call__(self, r):
        r.headers['Authorization'] = self.make_authorization_header_value()
        logger.debug(f"got remote services' jwt header: {r.headers}")
        return r

    def make_authorization_header_value(self, extra_payload: Optional[Dict] = None) -> str:
        """Make the value for "Authorization" header

        :param extra_payload: extra data which will be injected into token
        """
        payload = {
            'iss': self.auth_conf.iss,
            'expires_at': time.time() + self.expires_in,
            'role': self.auth_conf.role,
        }
        # Mix extra payload content
        payload.update(extra_payload or {})

        # use `six.ensure_str` to compatible with pyjwt 1.x
        token = six.ensure_str(jwt.encode(payload, key=self.auth_conf.key, algorithm=self.auth_conf.algorithm))
        return f'{self.prefix} {token}'



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
from decimal import Decimal

import six

_PROTECTED_TYPES = (
    type(None),
    int,
    float,
    Decimal,
)


def force_text(s, encoding='utf-8', strings_only=False, errors='strict'):
    """
    Similar to django's force_text function
    """
    # Handle the common case first for performance reasons.
    if issubclass(type(s), six.text_type):
        return s
    try:
        if not issubclass(type(s), six.string_types):
            if six.PY3:
                if isinstance(s, bytes):
                    s = six.text_type(s, encoding, errors)
                else:
                    s = six.text_type(s)
            elif hasattr(s, '__unicode__'):
                s = six.text_type(s)
            else:
                s = six.text_type(bytes(s), encoding, errors)
        else:
            # Note: We use .decode() here, instead of six.text_type(s, encoding,
            # errors), so that if s is a SafeBytes, it ends up being a
            # SafeText at the end.
            s = s.decode(encoding, errors)
    except UnicodeDecodeError:
        raise
    return s


def is_protected_type(obj):
    """Determine if the object instance is of a protected type.

    Objects of protected types are preserved as-is when passed to
    force_text(strings_only=True).
    """
    return isinstance(obj, _PROTECTED_TYPES)


def force_bytes(s, encoding='utf-8', strings_only=False, errors='strict'):
    """
    Similar to smart_bytes, except that lazy instances are resolved to
    strings, rather than kept as lazy objects.

    If strings_only is True, don't convert (some) non-string-like objects.
    """
    # Handle the common case first for performance reasons.
    if isinstance(s, bytes):
        if encoding == 'utf-8':
            return s
        else:
            return s.decode('utf-8', errors).encode(encoding, errors)
    if strings_only and is_protected_type(s):
        return s
    if isinstance(s, memoryview):
        return bytes(s)
    else:
        return s.encode(encoding, errors)



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import copy
import functools
from collections import namedtuple
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union

from django.conf import settings
from django.http.response import HttpResponseBase
from django.utils.module_loading import import_string
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.fields import empty
from rest_framework.serializers import BaseSerializer
from rest_framework.settings import api_settings
from rest_framework.utils.serializer_helpers import ReturnDict, ReturnList

if TYPE_CHECKING:
    from rest_framework.request import Request


def stringify_validation_error(error: ValidationError) -> List[str]:
    """Transform DRF's ValidationError into a list of error strings

    >>> stringify_validation_error(ValidationError({'foo': ErrorDetail('err')}))
    ['foo: err']
    """
    results: List[str] = []

    def traverse(err_detail: Any, keys: List[str]):
        """Traverse error data to collect all error messages"""

        # Dig deeper when structure is list or dict
        if isinstance(err_detail, (ReturnList, list, tuple)):
            for err in err_detail:
                traverse(err, keys)
        elif isinstance(err_detail, (ReturnDict, dict)):
            for key, err in err_detail.items():
                # Make a copy of keys so the inner loop won't affect outer scope
                _keys = copy.copy(keys)
                if key != api_settings.NON_FIELD_ERRORS_KEY:
                    _keys.append(str(key))
                traverse(err, _keys)
        else:
            if not keys:
                results.append(str(err_detail))
            else:
                results.append("{}: {}".format(".".join(keys), str(err_detail)))

    traverse(error.detail, [])
    return sorted(results)


#############
# drf crown #
#############
class WearOptions:
    is_unittest = False
    skip_swagger_schema = False


try:
    from drf_yasg.utils import swagger_auto_schema

except ImportError:
    WearOptions.skip_swagger_schema = True


ResponseParams = namedtuple("ResponseParams", "data,params")


_DEFAULT_SETTINGS_PREFIX = "DRF_CROWN_"


def enable_unittest():
    """Call me when you running testing"""
    WearOptions.is_unittest = True


@dataclass
class Config:
    """Config for Injector, control the process of injecting"""

    return_validated_data: bool = True
    remain_request: bool = False
    # sometime return raw data instead of serializer
    skip_out_cls: bool = False
    default_return_status: status = status.HTTP_200_OK


@dataclass
class ViewCrown:
    """A injector for injecting serializer as dependency"""

    body_in: Optional[Union[Type[BaseSerializer], BaseSerializer]]
    query_in: Optional[Union[Type[BaseSerializer], BaseSerializer]]
    out: Union[Type[BaseSerializer], BaseSerializer]
    config_params: Optional[dict] = field(default_factory=dict)
    valid_params: dict = field(default_factory=dict)

    def __post_init__(self):
        if self.query_in and self.body_in:
            raise ValueError("there should be only one param between in_body & in_query")

        self.valid_params = self.valid_params or {"raise_exception": True}

        # Priority decreases
        # 1. config as parameter from decorator
        # 2. config from django.settings
        # 3. config from Config class(above)
        _config = getattr(settings, _DEFAULT_SETTINGS_PREFIX + "DEFAULT_CONFIG", {}).copy()
        _config.update(self.config_params or {})
        self.config = Config(**_config)

        # remain an entrance for custom response class
        try:
            self.resp_cls = import_string(getattr(settings, _DEFAULT_SETTINGS_PREFIX + "RESP_CLS"))
        except AttributeError:
            self.resp_cls = import_string("rest_framework.response.Response")

    def get_in_serializer_instance(self, request: Optional["Request"] = None) -> "BaseSerializer":
        if not self.body_in and not self.query_in:
            raise ValueError("should given at least one serializer input")

        _data = empty
        if self.body_in:
            _in = self.body_in

            if request is not None:
                _data = getattr(request, "data")
        else:
            _in = self.query_in

            if request is not None:
                _data = getattr(request, "query_params")

        if isinstance(_in, BaseSerializer):
            # 由于传入的是全局对象，会残留上一次请求的结果
            # 这里需要手动清理一下
            if hasattr(_in, "_validated_data"):
                delattr(_in, "_validated_data")

            _in.initial_data = _data
            slz_obj = _in
        elif issubclass(_in, BaseSerializer):
            slz_obj = _in(data=_data)
        else:
            raise ValueError("unknown serializer input")

        return slz_obj

    def get_serializer_instance_by_request(self, request: "Request") -> "BaseSerializer":
        """Get in serializer instance"""
        slz_obj = self.get_in_serializer_instance(request)
        slz_obj.is_valid(**self.valid_params)
        return slz_obj

    def get_validated_data(self, request: "Request") -> dict:
        """Get validated data via in_serializer"""
        return self.get_serializer_instance_by_request(request).validated_data

    def get_in_params(self, request: "Request") -> dict:
        """Get extra params before view logic"""
        if WearOptions.is_unittest:
            return {}

        if self.config.return_validated_data:
            return {"validated_data": self.get_validated_data(request)}
        else:
            return {"serializer_instance": self.get_serializer_instance_by_request(request)}

    def get_response(self, data, out_params: dict) -> Any:
        """Get Response data"""
        if WearOptions.is_unittest:
            return data

        if self.config.skip_out_cls:
            return data

        if isinstance(data, (self.resp_cls, HttpResponseBase)):
            return data

        if isinstance(self.out, BaseSerializer):
            # 由于传入的是全局对象，会残留上一次请求的结果
            # 这里需要手动清理一下
            if hasattr(self.out, "_data"):
                delattr(self.out, "_data")

            self.out.instance = data
            _data = self.out.data
        elif issubclass(self.out, BaseSerializer):
            _data = self.out(data, **out_params).data
        else:
            raise ValueError("unknown serializer output")

        return self.resp_cls(data=_data, status=self.config.default_return_status)


def generate_swagger_params(crown: ViewCrown, swagger_params: dict) -> dict:
    """
    assemble params for swagger_auto_schema by crown
    """
    default_params = {}
    if crown.body_in:
        default_params = {"request_body": crown.get_in_serializer_instance()}
    elif crown.query_in:
        default_params = {"query_serializer": crown.get_in_serializer_instance()}

    if crown.out:
        default_params.update({"responses": {crown.config.default_return_status: crown.out}})

    default_params.update(swagger_params or {})
    return default_params


def inject_serializer(
    body_in: Optional[Union[Type[BaseSerializer], BaseSerializer]] = None,
    query_in: Optional[Union[Type[BaseSerializer], BaseSerializer]] = None,
    out: Optional[Union[Type[BaseSerializer], BaseSerializer]] = None,
    config: Optional[dict] = None,
    **swagger_kwargs
):
    """
    Sugar for simpling drf serializer specification
    :param body_in: input serializer (request body)
    :param query_in: input serializer (query)
    :param out: output serializer
    :param config: initial info of Config
    :param swagger_kwargs: pass to swagger_auto_schema of drf-yasg
    """

    def decorator_serializer_inject(func):
        crown = ViewCrown(body_in, query_in, out, config)

        if not WearOptions.skip_swagger_schema:
            func = swagger_auto_schema(**generate_swagger_params(crown, swagger_kwargs))(func)

        @functools.wraps(func)
        def decorated(*args, **kwargs):
            new_args = list(args)
            in_content: Dict[str, Any] = {}
            if body_in or query_in:
                in_content.update(**crown.get_in_params(new_args[1]))

            if not crown.config.remain_request:
                del new_args[1]

            original_data = func(*new_args, **kwargs, **in_content)
            if not out:
                return original_data

            # support runtime serializer params, like "context"
            params = {}
            if isinstance(original_data, ResponseParams):
                params = original_data.params
                original_data = original_data.data

            return crown.get_response(original_data, params)

        return decorated

    return decorator_serializer_inject



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
from typing import Any, Callable, Dict, Optional, Type, TypeVar, Union, overload

_DEFAULT_ERROR_CODE_NUM = -1
_DEFAULT_STATUS_CODE = 400

ExtraFormatterFunc = Callable[[str, 'APIError'], str]

T = TypeVar('T', bound='APIError')


class APIError(Exception):
    """API error object with detailed code and description

    :param code: required, An english identifier
    :param message: required, Detailed error message, may contains templated variables
    :param code_num: A numeric error code
    :param extra_formatter: an extra function for formatting message
    :param status_code: desired HTTP status code for representing current Error
    :param data: stores extra data in current Exception object
    """

    delimiter = ': '

    def __init__(
        self,
        code: str,
        message: str,
        code_num: int = _DEFAULT_ERROR_CODE_NUM,
        extra_formatter: Optional[ExtraFormatterFunc] = None,
        status_code: int = _DEFAULT_STATUS_CODE,
        data: Optional[Any] = None,
    ):
        self.code = code
        self.code_num = code_num
        self.extra_formatter = extra_formatter
        self.status_code = status_code
        self.data = data
        # Save message as private field to expose it as an property
        self._message = message

        super().__init__(self.message)

    def format(self: T, message: Optional[str] = None, replace: bool = False, **kwargs) -> T:
        """Try to mutate and render the original error message, return a cloned `APIError` object

        :param message: if not given, default message will be used
        :param replace: if True, relace the default message, otherwise append the incoming message
        """
        if not message:
            return self._clone(message=self._render(self._message, kwargs))

        if replace:
            obj_message = message
        else:
            # Note: use `f` to join str for compatibility with lazy loaded messages，such as django i18n gettext
            obj_message = f'{self._message}{self.delimiter}{message}'
        obj_message = self._render(obj_message, kwargs)
        return self._clone(message=obj_message)

    def set_data(self: T, data: Any) -> T:
        """A chain method which set data property"""
        self.data = data
        return self

    # Shortcut method name
    f = format

    @property
    def message(self) -> str:
        """Get detailed error message, it the `extra_formatter` was defined, it will be used for formatting"""
        if self.extra_formatter:
            return self.extra_formatter(self._message, self)
        return self._message

    def _clone(self: T, message: Optional[str] = None) -> T:
        """Clone a new APIError object

        :param message: if given, the cloned object will use this message instead of current `self._message`
        """
        obj_message = message if message is not None else self._message
        return self.__class__(
            code=self.code,
            code_num=self.code_num,
            extra_formatter=self.extra_formatter,
            status_code=self.status_code,
            data=self.data,
            message=obj_message,
        )

    @staticmethod
    def _render(message: str, kwargs: Dict) -> str:
        """Render message template with variables, using standard python string template syntax"""
        if kwargs:
            return message.format(**kwargs)
        return message

    def __str__(self) -> str:
        return f'<{self.__class__.__name__}: {self.code}({self.code_num})-{self.message}>'


class ErrorCode:
    """A descriptor object for defining error codes

    :param error_cls: the class for making exception instance, default to `APIError`
    """

    def __init__(self, *args, **kwargs):
        self._code = None
        self._error_cls = kwargs.pop('error_cls', APIError)
        # Save arguments for making error object later
        self._error_args = args
        self._error_kwargs = kwargs

    @overload
    def __get__(self, obj: None, obj_type: None) -> 'ErrorCode':
        ...

    @overload
    def __get__(self, obj: object, obj_type: Type) -> APIError:
        ...

    def __get__(self, obj: Union[None, object], obj_type: Union[None, Type]) -> Union['ErrorCode', APIError]:
        """When retriving `ErrorCode` via object attribute, always making a brand new `APIError`
        exception object
        """
        if obj is None:
            return self
        if not self._code:
            raise RuntimeError('must be initialized as a descriptor')

        try:
            return self._error_cls(self._code, *self._error_args, **self._error_kwargs)
        except TypeError as e:
            raise TypeError(f'Unable to make {self._error_cls.__name__} object: {e}')

    def __set_name__(self, obj_type, name):
        """Set field name as error code object's code"""
        self._code = name



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
from django.db import models

from blue_krill.encrypt.handler import EncryptHandler
from blue_krill.encrypt.utils import get_default_secret_key


class EncryptField(models.TextField):
    """a field which will be encrypted via cryptography/fernet"""

    description = "a field which will be encrypted"

    def __init__(self, secret_key: bytes = get_default_secret_key(), *args, **kwargs):
        super(EncryptField, self).__init__(*args, **kwargs)
        self.handler = EncryptHandler(secret_key=secret_key)

    def get_prep_value(self, value):
        if value is None:
            return value

        return self.handler.encrypt(value)

    def get_db_prep_value(self, value, connection, prepared=False):
        return self.get_prep_value(value)

    def from_db_value(self, value, expression, connection, context=None):
        if value is None:
            return value

        return self.handler.decrypt(value)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['blue_krill',
 'blue_krill.async_utils',
 'blue_krill.auth',
 'blue_krill.connections',
 'blue_krill.data_types',
 'blue_krill.editions',
 'blue_krill.encrypt',
 'blue_krill.models',
 'blue_krill.models.better_loaddata',
 'blue_krill.models.better_loaddata.management',
 'blue_krill.models.better_loaddata.management.commands',
 'blue_krill.monitoring',
 'blue_krill.monitoring.probe',
 'blue_krill.monitoring.prometheus',
 'blue_krill.redis_tools',
 'blue_krill.secure',
 'blue_krill.storages',
 'blue_krill.storages.blobstore',
 'blue_krill.web']

package_data = \
{'': ['*']}

install_requires = \
['click',
 'cryptography>=3.0.0,<4.0.0',
 'curlify>=2.2.1,<3.0.0',
 'django-environ>=0.8.1,<0.9.0',
 'pydantic>=1.7,<2.0',
 'pyjwt>=1.7.1',
 'python-editor>=1.0.4,<2.0.0',
 'requests',
 'six',
 'toml',
 'watchdog>=1.0.2,<2.0.0']

extras_require = \
{':python_full_version >= "3.6.2" and python_version < "3.7"': ['dataclasses>=0.7,<0.8']}

entry_points = \
{'console_scripts': ['bk-secure = blue_krill.secure.bk_secure:main',
                     'editionctl = blue_krill.editions.editionctl:main']}

setup_kwargs = {
    'name': 'blue-krill',
    'version': '1.2.2',
    'description': 'Tools and common packages for blueking paas',
    'long_description': '# Blue krill\n\nPython 常用工具包模块。\n\n## 使用指南\n\n### 1. blue_krill.secure\n\n加解密、安全相关的工具与模块。\n\n#### 1.1 dj_environ 模块\n\ndj_environ 的主要目的是在 [django-environ](https://github.com/joke2k/django-environ) 之上，增加可以阅读加密环境变量的能力。如需使用，首先需要设置密文环境变量：\n\n```bash\nexport DATABASE_URL=\'<使用 bk-secure 加密后的字符串>\'\nexport FOO=\'<使用 bk-secure 加密后的字符串>\'\n```\n\n然后在 `settings` 模块中，初始化 `SecureEnv` 对象并使用它读取对应配置：\n\n```python\nfrom blue_krill.secure.dj_environ import SecureEnv\n\n# 初始化并加载 .env 文件内容\nsec_env = SecureEnv()\nenviron.Env.read_env()\n\n# 读取为数据库配置\nDATABASES[\'default\'] = sec_env.db()\n\n# 读取为普通配置\nFOO = sec_env(\'FOO\')\n```\n\n#### 1.2 bk-secure 脚本\n\nbk-secure 主要用于配合 dj_environ 模块生成加密环境变量（或配置文件）。使用前，先将 `BK_FERNET_KEY` 设为你所使用的加密 key。一般情况下，这个值等同于 Django 项目的 `BKKRILL_ENCRYPT_SECRET_KEY` 配置项：\n\n```\nexport BK_FERNET_KEY=\'... ...\'\n```\n\n执行 `encrypt` 加密某段明文：\n\n```console\n❯ bk-secure encrypt\nInput string: mysql://u:p@localhost/foo\nThe encrypted token is: gAAAAABfKUtKIBzYc_gyQL-j9TmI35O1d0auLQfYeso6D8Q77ZC9PIuv26ABPFlOQSSPDzT3HcVrhI1K3XwU5Xfs6gP6iAe8RhEAJJhMktp7CKzn7p7imNk=\n```\n\n执行 `decrypt` 解密某段密文：\n\n```console\n❯ bk-secure decrypt\nInput token: gAAAAABfKUtKIBzYc_gyQL-j9TmI35O1d0auLQfYeso6D8Q77ZC9PIuv26ABPFlOQSSPDzT3HcVrhI1K3XwU5Xfs6gP6iAe8RhEAJJhMktp7CKzn7p7imNk=\nThe decrypted result is: mysql://u:p@localhost/foo\n```\n\n使用 `edit` 以明文方式编辑某段密文，并输出新的密文：\n\n```console\n❯ bk-secure edit\nInput token: gAAAAABfKUtKIBzYc_gyQL-j9TmI35O1d0auLQfYeso6D8Q77ZC9PIuv26ABPFlOQSSPDzT3HcVrhI1K3XwU5Xfs6gP6iAe8RhEAJJhMktp7CKzn7p7imNk=\n# 将弹出编辑器界面修改明文，可通过 $EDITOR 环境变量设置编辑器\nThe new value is: mysql://u:p@localhost/foo2\nThe encrypted new value is: gAAAAABfKUui5_YUVxoYEYQG61RSRX1Ll3s1dgkZ5nUEJbCxakWHSyo3AKZFv3GuoQ7cH2Hm5LEU2QDK8C3G-_iog0TmqSbVkIYf0WnksH2DGgedldfbwhs=\n```\n\n### 2. blue_krill.editions\n\n多版本开发相关工具模块\n\n#### 2.1 editionctl 工具\n\n`editonctl` 是用来帮助开发需要支持多版本 Python 项目的工具，使用该工具前，请先把你的项目目录组织成下面这种结构：\n\n```raw\n├── editionctl.toml\n├── editions\n│\xa0\xa0 ├── ee\n│\xa0\xa0 │\xa0\xa0 └── ee.py\n│\xa0\xa0 └── te\n│\xa0\xa0     └── te.py\n└── main\n    └── main.py\n```\n\n其中：\n\n- `editions`：仅保存不同版本所特有的源码文件\n- `main`：项目主目录\n\n#### 2.1.1 创建配置文件\n\n要使用 `editonctl`，首先需要在项目内创建配置文件 `editionctl.toml`。比如，针对上面的项目结构，我们可以创建这样的配置文件：\n\n```toml\n# 项目主目录\nproject_root = \'main\'\n# 项目各版本所在目录\neditions_root = \'editions\'\n\n[[editions]]\n# 版本名称\nname = "TE"\n# 版本相对路径\nrel_directory = \'te\'\n\n[[editions]]\nname = "EE"\nrel_directory = \'ee\'\n```\n\n> 更多配置文件相关说明，可执行 `editionctl help` 查看。\n\n#### 2.1.2 在版本之间切换\n\n在不同版本间切换，需要使用 `editionctl activate {EDITION_NAME}` 命令。执行该命令后，工具会将指定版本下的所有源码文件，拷贝到 `project_root` 中。\n\n```bash\n$ editionctl activate EE\n[2020-12-17 16:51:37,312] INFO: Edition EE activated, linker is default\n```\n\n> 为了避免由工具拷贝的文件被意外提交到源码仓库，这些文件会被添加到 `{project_root}/.gitignore` 中。\n\n#### 2.1.3 重置多版本\n\n假如你想要清除所有由 `editionctl` 工具创建的源码文件，可以执行 `editionctl reset` 命令。执行该命令将删除所有多版本相关文件，只保留主目录。\n\n#### 2.1.4 进入开发模式\n\n在开发多版本项目时，`editions_root` 目录下的当前版本相关文件会被频繁修改。正常情况下，每次修改版本文件后，我们都要手动重新执行 `activate` 命令重新同步文件。\n\n为了简化这个过程，我们可以使用 `editionctl develop`。\n\n执行 `editionctl develop` 命令后，工具将会持续监听当前 `edition` 目录下的任何改动。如果监听到新改动，则自动触发同步机制。\n\n```\n$ editionctl develop\n[2020-12-17 16:56:34,385] INFO: Start watching editions/ee directory for edition EE...\n```\n\n### 3. blue_krill.data_types.enum\n\n枚举相关的数据类型。\n\n### 3.1 FeatureFlag\n\n功能标记(Feature Flag)用于控制当前用户能否感知到某个功能/特性，只提供**开启(enabled)**和**关闭(disabled)**两个状态, 分别对应于布尔值的 True/Flase。\n为了避免各项目重复造轮子, blue_krill 抽象出通用的 FeatureFlag 模型, 同时也提供类似于`枚举(Enum)`的API, 降低使用成本。\n\n#### 3.1.1 如何定义 FeatureFlag\n\n就像定义普通的 Python Class 一样, 定义 FeatureFlag 只需要继承 `blue_krill.data_types.enum::FeatureFlag` 即可。\n```python\nfrom blue_krill.data_types.enum import FeatureFlag, FeatureFlagField\n\n\nclass UserFeatureFlag(FeatureFlag):\n    # 使用类属性声明 FeatureFlagField 时, name 属性会通过描述符协议自动设置, 无需额外指定.\n    WEBCONSOLE = FeatureFlagField(label="使用 WEBConsole")\n    CREATE_SMART_APP = FeatureFlagField(label="创建 Smart 应用")\n    ...\n```\n\n#### 3.1.2 如何添加额外的 FeatureFlag\n\n不同于`枚举值(Enum)`, FeatureFlag 允许在运行过程中动态添加额外的字段或修改已有字段的默认值。\n\n```python\n# 添加额外的 FeatureFlag 时, 需要制定对应的名称.\nUserFeatureFlag.register_feature_flag(FeatureFlagField(name="CHOOSE_SOURCE_ORIGIN", label="选择源码来源"))\n```\n\n### 3.2 StructuredEnum\n\n考虑到我们使用`枚举值(Enum)`时, 往往会给枚举值添加额外的描述字段，为了避免各项目重复造轮子，blue_krill 基于 `Enum` 实现了 StructuredEnum，可以基于配套的 `EnumField` 定义带有额外描述内容的枚举值。\n\n```python\nfrom blue_krill.data_types.enum import EnumField, StructuredEnum\n\n\nclass DiffType(str, StructuredEnum):\n    ADDED = EnumField("added", label="新增")\n    DELETED = EnumField("deleted", label="删除")\n    NOT_MODIFIED = EnumField("not_modified", label="未改动")\n```\n\n### 4. blue_krill.storages.blobstore\n\n对象存储服务的简单封装, 目前支持 **上传**, **下载**, **生成预签名URL** 三个接口.\n\n#### 4.1 S3Store\n\nS3 协议的 BlobStore 实现, 使用时需要额外安装 boto3==\'^1.4.3\', 可参考以下代码进行实例化:\n```python\nfrom blue_krill.storages.blobstore.s3 import S3Store\n\n\nstore = S3Store(\n    bucket="your-bucket",\n    aws_access_key_id=\'your-access-key\',\n    aws_secret_access_key=\'your-secret-key\',\n    endpoint_url=\'your-s3-endpoint\',\n    # Optional\n    region_name=\'your-region, default is `us-east-1`\',\n    signature_version=\'your-signature-version, default is s3v4\',\n)\n```\n\n#### 4.2 BKGenericRepo\n\n底层服务是 **蓝鲸通用二进制仓库** 的底层实现, 可参考以下代码进行实例化:\n\n```python\nfrom blue_krill.storages.blobstore.bkrepo import BKGenericRepo\n\n\nstore = BKGenericRepo(\n    bucket=\'your-bucket\',\n    project=\'your-project-id\',\n    endpoint_url=\'\',\n)\n```\n\n### 5. blue_krill.web\n\n`blue_krill.web` 主要存放与 Web 开发有关的工具集。\n\n#### 5.1 blue_krill.web.std_error\n\n该模块内包含标准的错误码功能。`std_error` 最常见的用法，是通过 `ErrorCode` 定义一套错误码集合：\n\n```python\nfrom blue_krill.web.std_error import ErrorCode\n\nclass ErrorCodes:\n    CREATE_ERROR = ErrorCode(\'创建失败\')\n    DELETE_ERROR = ErrorCode(\'删除失败\')\n\n# 实例化一个全局对象\nerror_codes = ErrorCodes()\n```\n\n当你要抛出某个特定错误时，可以使用下面的语句：\n\n```python\nraise error_codes.CREATE_ERROR\n\n# 使用 .f() / .format() 方法追加错误信息\nraise error_codes.CREATE_ERROR.f(\'追加说明\')\n\n# 传递 replace=True 替换错误信息\nraise error_codes.CREATE_ERROR.f(\'替换信息\', replace=True)\n\n# 设置异常对象的 data 属性，以便在后续处理。可通过 exc_obj.data 属性读取\nraise error_codes.CREATE_ERROR.set_data({\'your\': \'data\'})\n```\n\n> 注意：`APIError` 是不可变类型，调用 `format()` 会克隆并返回一个新对象，而非修改现有对象。\n\n当程序抛出 `APIError` 异常后，为了让用户正常看到错误响应，你必须在 Web 框架里捕获并处理该类异常。\n\n##### 5.1.1 在框架内捕获 APIError 异常\n\n视 Web 框架的不同，捕获与处理 `APIError` 的方式会略有区别，以 `DRF` 框架为例。要捕获 `APIError` 异常，我们首先得创建一个新函数：\n\n```python\n# file: my_module.py\nfrom blue_krill.web.std_error import APIError\n\ndef custom_exception_handler(exc, context):\n    if isinstance(exc, APIError):\n        # 你可以随意修改这里的响应数据结构\n        data = {\n            \'code\': exc.code,\n            \'detail\': exc.message,\n        }\n        return Response(data, status=exc.status_code)\n    # ... 其他异常处理撮箕\n```\n\n创建完函数后，下一步是修改项目配置，将 `EXCEPTION_HANDLER` 调整为该异常捕获函数：\n\n```python\nREST_FRAMEWORK = {\n    \'EXCEPTION_HANDLER\': \'my_module.custom_exception_handler\'\n}\n```\n\n配置完成后，每当你在项目里抛出 `error_codes.CREATE_ERROR`，用户便会看到下面的错误信息：\n\n```json\n{\n    "code": \'CREATE_ERROR\',\n    "detail": \'创建失败\'\n}\n```\n\n##### 5.1.2 ErrorCode API 说明\n\n在创建 `ErrorCode` 对象时，除了可以传入必须的 `message` 参数，还支持传入许多可选的个性化参数：\n\n- `message`: 必选，错误详情信息，可包含字符串模板变量\n- `code_num`: 可选，数字错误代码，默认为 -1\n- `extra_formatter`：可选，额外的错误信息格式化函数\n- `status_code`: 该错误推荐使用的 HTTP 错误代码，默认为 400\n\n这些参数各自有着不同的用途，比如，通过定义 `extra_formatter` 属性，你可以调整 `APIError` 拼装错误信息 `message` 的逻辑。\n\n以下面的代码为例：\n\n```python\n# formatter 函数接收两个参数：默认错误信息、当前异常对象\ndef _format_message(message, exc):\n    # 将错误码拼装到错误信息前\n    return f\'code: {exc.code} - {message}\'\n\nclass ErrorCodes:\n    foo = ErrorCode(\'foo message\', extra_formatter=_format_message)\n```\n\n当你抛出 `foo` 错误码时，由于我们使用了自定义的 `message` 格式化函数，错误详情会变成这样：`code: foo - foo message`。\n\n`ErrorCode` 的 `message` 除了能使用普通字符串以外，还支持字符串模板功能。举个例子，假如你定义的 `message` 是 `name={name}`，那么，当你抛出异常时，可以用 `.format()` 方法传入模板变量，对错误信息进行二次渲染。\n\n```python\nraise error_codes.FOO # 1\nraise error_codes.FOO.format(name=\'foobar\') # 2\n```\n\n1. 用户看到的错误信息是 `name={name}`\n2. 用户看到的错误信息会变为 `name=foobar`\n\n#### 5.2 blue_krill.web.drf_utils\n\n`drf_utils` 模块内包含许多与 DRF 框架有关的工具。\n\n##### 5.2.1 stringify_validation_error\n\n`stringify_validation_error()` 会将由 DRF 框架抛出的 `ValidationError` 校验错误异常对象，转换为可读性更好的错误提示文字。\n\n比如，下面的异常对象：\n\n```python\nValidationError({\'foo\': {\'bar\': [ErrorDetail(\'err1\'), ErrorDetail(\'err2\')]}})\n```\n\n可以被转换为：`[\'foo.bar: err1\', \'foo.bar: err2\']` 这样的文字内容。\n\n### 6. blue_krill.async_utils\n\n#### 6.1 blue_krill.aysnc_utils.poll_tasks\n\n`poll_tasks` 是一个用来执行长时间轮询任务的异步工具模块。它的工作原理是每隔几秒钟，拉起一个 `celery` 任务进行轮询逻辑。当轮询应该结束时，带着结果回调。\n\n要创建一个新的轮询任务，你首先要编写一个 `TaskPoller` 类。\n\n```python\nfrom blue_krill.async_utils.poll_task import TaskPoller, PollingResult\n\nclass MyTaskPoller(TaskPoller):\n\n    # 通过定义下面的属性，修改当前 Poller 类的默认配置\n    # max_retries_on_error = 10\n    # overall_timeout_seconds = 3600 * 24 * 7\n    # default_retry_delay_seconds = 10\n\n    def query(self) -> PollingResult:\n        result = request_api()\n        if result:\n            return PollingResult.done(data={\'result\': ...})\n        else:\n            return PollingResult.doing(data={\'current_value\': ...})\n```\n\n`TaskPoller` 的配置属性含义如下：\n\n- `max_retries_on_error`：当轮询抛出异常的总次数，超过该值后，不再继续下次轮询\n- `overall_timeout_seconds`：当轮询的总执行时间（从第一次轮询开始后计算）超过该值，结束本次轮询并返回超时结果\n- `default_retry_delay_seconds`：两次轮询行为之间相隔的秒数\n\n每个 `TaskPoller` 类都必须重写 `query()` 方法，在其中实现**每次**轮询的真正逻辑。在 `query()` 方法内部，你可以从以下属性读取与本次轮询相关的数据：\n\n- `self.params`：轮询任务启动时的参数，通常为字典 `Dict`\n- `self.metadata`：本次轮询任务的元数据，里面包含轮询开始时间、已完成的轮询次数等数据\n\n`query()` 方法需要返回一个 `PollingResult` 结果，来控制接下来的轮询流程。\n\n不同的轮询结果，代表着不同含义：\n\n- `PollingResult.done()`：表示整个轮询任务已结束，不会启动新异步任务\n- `PollingResult.doing()`：表示应该继续轮询，会派生新的异步任务\n\n在实例化 `PollingResult` 时，你可以通过 `data` 属性传入额外数据。该数据对于不同状态的轮询结果来说，有着不同含义。\n\n- 当轮询返回 `done()` 结果时，`data` 会通过回调传递到 `CallbackHandler` 的 `result` 参数里\n- 当轮询仍在继续，返回 `doing()` 结果时，可在 `TaskPoller` 类中，通过 `self.metadata.last_polling_data` 获取**上次轮询**的 `data` 内容\n\n创建完 TaskPoller 类后，下一步是编写 ResultHandler 结果回调类。\n\n##### 6.1.1 定义 CallbackHandler 类\n\n一个标准的 `CallbackHandler` 如下所示：\n\n```python\nfrom blue_krill.async_utils.poll_task import (\n    CallbackHandler,\n    CallbackResult,\n    TaskPoller\n)\n\nclass MyHandler(CallbackHandler):\n\n    def handle(self, result: CallbackResult, poller: TaskPoller):\n        # 通过 result 和 poller 执行回调逻辑\n        pass\n```\n\n根据轮询的不同执行结果，`CallbackResult.status` 会有几种不同的状态：\n\n- `CallbackStatus.NORMAL`：轮询正常结束，`Poller` 返回的轮询结果为 `DONE` / \'.doing()\'\n- `CallbackStatus.TIMEOUT`：轮询超过了规定时间（``overall_timeout_seconds``）仍未结束，判定为超时\n- `CallbackStatus.EXCEPTION`：轮询时发生了异常，且总异常次数超过最大值：`max_retries_on_error`\n\n为了方便操作，你可以直接调用 `result.is_exception` 属性来获知本次轮询是否正常结束。任何状态不为 `NORMAL` 的结果，`.is_exception` 值都为 `True`。\n\n##### 6.1.2 启动轮询任务\n\n当你定义好 `Poller` 与 `CallbackHandler` 类后，可以用以下方式启动一次轮询任务：\n\n```python\n# params = {\'some_field\': \'value\'}\nMyTaskPoller.start(params, MyHandler)\n```\n\n通过执行 `TaskPoller` 类的 `start()` 方法，程序会派生出一个名为 `poll_task.check_status_until_finished` 的 `celery` 异步任务，之后触发 `TaskPoller` 的 `query()` 方法，不断开始轮询。\n\n### 7. blue_krill.monitoring.probe\n\n`blue_krill.monitoring.probe` 模块提供了常见的健康探针功能。\n\n#### 7.1 blue_krill.monitoring.probe.tcp\n\n`blue_krill.monitoring.probe.tcp` 模块提供了通用的 TCP 健康探针, 可检测是否能建立 TCP 连接。\n\n```python\n# Usage:\nfrom blue_krill.monitoring.probe.tcp import TCPProbe, InternetAddress\n\n\nclass SomeTCPProbe(TCPProbe):\n    name: str = "some"\n    address: InternetAddress = InternetAddress(host="localhost", port=8080)\n\n\nreport = SomeTCPProbe().report()\n```\n\n#### 7.2 blue_krill.monitoring.probe.http\n\n`blue_krill.monitoring.probe.http` 模块提供了通用的 HTTP 健康探针, 可检测 HTTP 服务是否正常工作。\n\n```python\n# Usage\nfrom blue_krill.monitoring.probe.http import HttpProbe\n\n\nclass SomeHttpProbe(HttpProbe):\n    name: str = "some"\n    url: str = "http://localhost/ping"\n\n\nreport = SomeHttpProbe().report()\n\n\nclass SomeHttpWithAuth(HttpProbe):\n    name: str = "some"\n    url: str = "http://localhost/ping"\n    params: Dict = {"token": "dummy"}\n    headers: Dict = {"Authorization": "Basic YWxhZGRpbjpvcGVuc2VzYW1l"}\n\n\nreport = SomeHttpWithAuth().report()\n```\n\n#### 7.3 blue_krill.monitoring.probe.mysql\n\n`blue_krill.monitoring.probe.mysql` 模块提供了通用的 MySQL 健康探针, 可检测 MySQL 服务是否正常工作, 该模块依赖 pymysql。\n\n```python\n# Usage:\nfrom blue_krill.monitoring.probe.mysql import MySQLProbe, MySQLConfig\n\n\nclass SomeMySQLProbe(MySQLProbe):\n    name: str = "some"\n    config = MySQLConfig(host="localhost", port=3306, username="root", password="root", database="information_schema")\n\n\nreport = SomeMySQLProbe().report()\n```\n\n#### 7.4 blue_krill.monitoring.probe.redis\n\n`blue_krill.monitoring.probe.redis` 模块提供了通用的 Redis 健康探针和 Redis Sentinel 集群健康探针, 可检测 Redis 服务是否正常工作, 该模块依赖 redis。\n\n```python\n# Usage:\nfrom blue_krill.monitoring.probe.redis import RedisProbe\n\n\nclass SomeRedisProbe(RedisProbe):\n    name: str = "some"\n    redis_url: str = "redis://localhost:6379/0"\n\n\nreport = SomeRedisProbe().report()\n\n\n# Redis Sentinel\nfrom blue_krill.monitoring.probe.redis import RedisSentinelProbe\n\n\nclass SomeRedisSentinelProbe(RedisSentinelProbe):\n    name: str = "some"\n    redis_url: str = "sentinel://:xxx@localhost:6379/0"\n    master_name: str = "mycluster"\n    sentinel_kwargs: dict = {\'password\': \'xxx\'}\n\n\nreport = SomeRedisSentinelProbe().report()\n```\n\n### 8 blue_krill.cubing_case\n`blue_krill.cubing_case` 增加各个方法互相转换的工具库.\n\n#### 8.1 blue_krill.cubing_case.RegexCubingHelper\n基于多种正则将多种模式混合的字符串进行拆分，转换并组合成新的字符串的工具类。\n\n#### 8.2 blue_krill.cubing_case.CommonCaseConvertor\n在 `blue_krill.cubing_case.RegexCubingHelper` 之上的一个封装实现，将指定的多种模式的字符串转化成常见的方法，包含：\n- 驼峰式：`CubingCase`\n- 小写开头的驼峰式：`cubingCase`\n- 小写下划线式：`cubing_case`\n- 大写下划线式：`CUBING_CASE`\n- 小写连字符式：`cubing-case`\n- 大写下划线式：`CUBING-CASE`\n- 小写点分式：`cubing.case`\n- 大写下划线式：`CUBING.CASE`\n- 小写空格分隔式：`cubing case`\n\n#### 8.3 blue_krill.cubing_case.shortcuts\n`blue_krill.cubing_case.shortcuts` 是 `blue_krill.cubing_case.CommonCaseConvertor` 的一个快捷方式，内置了其转换目标的所有源模式，可以实现所有模式的正反转换。\n\n### 9 blue_krill.redis_tools\n`blue_krill.redis_tools` 提供了 redis 常用工具\n\n#### 9.1 blue_krill.redis_tools.sentinel\n`blue_krill.redis_tools.sentinel` 提供了 redis sentinel 模式下，直接从 url 生成 redis 实例的方法。\n\n```python\n# Usage:\nfrom blue_krill.redis_tools.sentinel import SentinelBackend\n\n\nbackend = SentinelBackend(\'sentinel://xxx@localhost:26347/0\', \'mycluster\', {\'password\': \'xxx\'})\nr = backend.client\n\nr.set(\'foo\', \'bar\')\n# 获得 b\'bar\'\nr.get(\'foo\')\n```\n\n## 开发指南\n\n首先安装 [poetry](https://github.com/python-poetry/poetry)，之后在项目目录下执行 `poetry env use python3.6` 初始化开发用虚拟环境。然后用 `poetry shell` 命令激活虚拟环境。\n\n- 执行 `poetry install` 安装所有依赖\n- 使用 `pytest -s .` 执行所有单元测试\n\n在开发时，如果想让某项目安装本地目录里的 blue-krill 模块，首先切换到对应项目虚拟环境，然后在 blue-krill 目录执行 `pip install -e .` \n\n### 使用 tox 执行单元测试\n\n为了测试包在不同 Python 版本下的稳定性，我们使用了 tox 工具。在项目目录下执行 `tox` 即可执行所有的单元测试。\n\n### 发布包\n\n首先，执行 `poetry build` 命令在 dist 目录下生成当前版本的包(需要检查 dist 目录中的内容是否符合预期，避免上传其他版本覆盖)。然后执行 `twine upload dist/* --repository-url {pypi_address} --username {your_name} --password {your_token}` 将其上传到 pypi 服务器上。\n\n### 以“可编辑模式”安装\n\n进入项目根目录，执行 `pip install -e .` 以“可编辑模式”安装包。该操作要求 pip 为版本 21.3 及以上[（参考）](https://pip.pypa.io/en/stable/news/#v21-3)。\n',
    'author': 'blueking',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<3.11',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import logging
from os import PathLike
from shutil import copyfileobj
from typing import BinaryIO

from blue_krill.storages.blobstore.base import BlobStore, SignatureType
from blue_krill.storages.blobstore.exceptions import DownloadFailedError, ObjectAlreadyExists, UploadFailedError

try:
    import boto3
    from botocore.client import Config
    from botocore.exceptions import ClientError
except ImportError as e:
    raise ImportError("Please install boto3=='^1.4.3'") from e


logger = logging.getLogger(__name__)


class S3Store(BlobStore):
    """支持 S3 协议的对象存储服务"""

    STORE_TYPE = "s3"

    def __init__(
        self,
        bucket: str,
        aws_access_key_id: str,
        aws_secret_access_key: str,
        endpoint_url: str,
        region_name: str = 'us-east-1',
        signature_version: str = "s3v4",
        **kwargs,
    ):
        super().__init__(bucket)
        self.signature_version = signature_version
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.endpoint_url = endpoint_url
        self.region_name = region_name

    def get_client(self, signature_version: str = None):
        return boto3.resource(
            "s3",
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            endpoint_url=self.endpoint_url,
            config=Config(
                region_name=self.region_name,
                signature_version=signature_version or self.signature_version,
            ),
        )

    def upload_file(self, filepath: PathLike, key: str, allow_overwrite: bool = True, **kwargs):
        client = self.get_client()
        logger.info("Start uploading package to S3 server...")
        bucket = client.Bucket(self.bucket)
        # 由于 S3 默认的上传行为是覆盖, 因此如果不允许覆盖, 只能先发请求判断文件是否存在, 再进行上传操作.
        if not allow_overwrite:
            self.check_key_exists_in_bucket(client, bucket=self.bucket, key=key)

        try:
            bucket.upload_file(str(filepath), key, **kwargs)
            logger.info("Upload to '%s' success.", key)
        except Exception as e:
            logger.exception("upload to '%s' failed.", key)
            raise UploadFailedError(key=key, src=str(filepath)) from e

    def upload_fileobj(self, fh: BinaryIO, key: str, allow_overwrite: bool = True, **kwargs):
        client = self.get_client()
        logger.info("Start uploading package to S3 server...")
        bucket = client.Bucket(self.bucket)
        # 由于 S3 默认的上传行为是覆盖, 因此如果不允许覆盖, 只能先发请求判断文件是否存在, 再进行上传操作.
        if not allow_overwrite:
            self.check_key_exists_in_bucket(client, bucket=self.bucket, key=key)

        try:
            bucket.upload_fileobj(fh, key, **kwargs)
            logger.info("Upload to '%s' success.", key)
        except Exception as e:
            logger.exception("upload to '%s' failed.", key)
            raise UploadFailedError(key=key, src='<memory>') from e

    def download_file(self, key: str, filepath: PathLike, *args, **kwargs) -> PathLike:
        client = self.get_client()
        bucket = client.Bucket(self.bucket)
        try:
            bucket.download_file(key, str(filepath))
        except Exception as e:
            raise DownloadFailedError(key=key, dest=str(filepath)) from e
        return filepath

    def download_fileobj(self, key: str, fh: BinaryIO, *args, **kwargs):
        client = self.get_client()
        bucket = client.Bucket(self.bucket)
        try:
            obj = bucket.Object(key).get()
            copyfileobj(obj["Body"], fh)
        except Exception as e:
            raise DownloadFailedError(key=key, dest='<memory>') from e

    def delete_file(self, key: str, *args, **kwargs):
        client = self.get_client()
        bucket = client.Bucket(self.bucket)
        return bucket.Object(key).delete()

    def get_file_metadata(self, key: str):
        client = self.get_client()
        bucket = client.Bucket(self.bucket)
        obj = bucket.Object(key).get()
        return obj

    def generate_presigned_url(
        self, key: str, expires_in: int, signature_type: SignatureType = SignatureType.DOWNLOAD, *args, **kwargs
    ) -> str:
        # Must set signature_version to s3 instead of s3v4 or the generated presigned url
        # won't work "signature not match" error
        operation_name = {SignatureType.DOWNLOAD: "get_object", SignatureType.UPLOAD: "put_object"}[
            (SignatureType(signature_type))
        ]
        client = self.get_client(signature_version="s3")
        url = client.meta.client.generate_presigned_url(
            operation_name,
            Params={"Bucket": self.bucket, "Key": key},
            ExpiresIn=expires_in,
        )
        return url

    @staticmethod
    def check_key_exists_in_bucket(client, bucket, key):
        try:
            client.Object(bucket, key).load()
            raise ObjectAlreadyExists(f"A source package with the same key[{key}] already exists.", code="Unknown")
        except ClientError as e:
            code = e.response.get("Error", {}).get("Code")
            if code != "404":
                raise ObjectAlreadyExists(
                    f"A source package with the same key[{key}] already exists.", code=code, response=e.response
                )



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import datetime
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .ha_endpoint_pool import Endpoint


class BasicHAAlgorithm:
    """Algorithm class which cooperates with `HAEndpointPool`, it decides how
    Endpoint recover or become unhealthy.
    """

    # Max attempts before consider endpoint as "unhealthy"
    unhealthy_max_failed: int = 3
    # Unhealthy Endpoint become healthy again after `unhealthy_period`
    unhealthy_period: datetime.timedelta = datetime.timedelta(seconds=20)

    def is_unhealthy(self, endpoint: 'Endpoint') -> bool:
        """Whether given endpoint is unhealthy"""
        return endpoint.failure_count >= self.unhealthy_max_failed

    def should_recover(self, endpoint: 'Endpoint') -> bool:
        """Whether an unhealthy endpoint should be recovered"""
        if not endpoint.unhealthy_at:
            return False
        return (datetime.datetime.now() - endpoint.unhealthy_at) >= self.unhealthy_period



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import logging
import time
from typing import Dict, Optional, Tuple

import jwt
from django.http import HttpRequest
from jwt.exceptions import DecodeError

from .jwt import DEFAULT_ALGORITHM
from .utils import get_paas_service_jwt_clients, validate_jwt_token

logger = logging.getLogger(__name__)


class AuthFailedError(Exception):
    """Raised when one authenticator fails"""


class Client:
    def __init__(self, name, auth_backend_type=None, role=None):
        self.name = name
        self.auth_backend_type = auth_backend_type
        self.role = role

    def __str__(self):
        return self.name

    def is_verified(self):
        return True

    @classmethod
    def from_jwt_settings(cls, data):
        for key in ('iss', 'key'):
            if not data.get(key):
                raise ValueError(f'error: {key} is required')
        return cls(name=data['iss'], auth_backend_type='jwt')


class AuthResult:
    """result for authentication"""

    def __init__(self, client: Client, extra_payload: Dict):
        self.client = client
        self.extra_payload = extra_payload


class JWTClientAuthenticator:
    """Authenticator using JWT clients"""

    def __init__(self, jwt_clients):
        self.jwt_clients = jwt_clients

    def authenticate(self, token: str) -> AuthResult:
        """Authenticate a given token

        :raises: AuthFailedError
        """
        try:
            payload, client = self.parse(token)
        except ValueError as e:
            logger.warning(f'authentication failed, token is invalid: {e}')
            raise AuthFailedError('token is invalid')
        else:
            logger.info(f'successfully authenticated token issued by {client}')
            return AuthResult(client=client, extra_payload=payload)

    def parse(self, token: str) -> Tuple[Dict, Client]:
        """Parse given token

        :raises: ValueError when token is invalid
        """
        for client in self.jwt_clients:
            try:
                algorithm = client.get('algorithm', DEFAULT_ALGORITHM)
                payload = jwt.decode(token, client['key'], algorithms=[algorithm])
            except DecodeError:
                logger.debug(f'Unable to decode token using {client["iss"]}\'s secret')
                continue

            if not self._validate_payload(client, payload):
                logger.debug('payload validation failed')
                continue

            client_ins = Client.from_jwt_settings(client)
            client_ins.role = payload.get('role')
            return payload, client_ins
        raise ValueError('invalid JWT token')

    @staticmethod
    def _validate_payload(client: Dict, payload: Dict) -> bool:
        """Validates given JWT payload, a valid payload must contains at least 2 fields:
        'iss' and 'expires_at'.
        """
        expires_at = payload.get('expires_at')
        try:
            expires_at = int(expires_at)  # type: ignore
        except (TypeError, ValueError):
            logger.warning("token's expires_at is empty or invalid")
            return False
        if expires_at < time.time():
            logger.warning("token has already expired")
            return False

        iss = payload.get('iss')
        if not (iss and iss == client['iss']):
            logger.warning("issuer name in payload does not match client's config")
            return False
        return True


class VerifiedClientMiddleware:
    """when current request has a valid signed JWT token in headers, middleware will set two
    property on `request` object:

    - request.client: verified client object
    - request.extra_payload: extra payload from JWT token, usually includes current username etc.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    @staticmethod
    def get_token(request) -> Optional[str]:
        """Get token from request, return None if no token can be found"""
        auth_header_val = request.headers.get('Authorization') or request.META.get('HTTP_AUTHORIZATION')
        if not auth_header_val:
            return None

        try:
            _, token = auth_header_val.split(None, 1)
        except ValueError:
            logger.warning('Invalid authorization header value')
            return None
        return token

    def __call__(self, request):
        token = self.get_token(request)
        request.client = None
        request.extra_payload = None
        # Only proceed when token format is valid JWT format
        if token and validate_jwt_token(token):
            try:
                ret = JWTClientAuthenticator(get_paas_service_jwt_clients()).authenticate(token=token)
                request.client = ret.client
                request.extra_payload = ret.extra_payload
            except AuthFailedError as e:
                logger.info('Unable to create an authenticated client using token: %s, reason: %s', token, e)

        response = self.get_response(request)
        return response


def check_client_role(request: HttpRequest, role: str) -> bool:
    """Check if request contains a verified client and it's role matches given value"""
    client = getattr(request, 'client', None)
    if not (client and client.is_verified()):
        return False
    return request.client.role == role



# -*- coding: utf-8 -*-
"""
 * TencentBlueKing is pleased to support the open source community by making 蓝鲸智云-蓝鲸 PaaS 平台(BlueKing-PaaS) available.
 * Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://opensource.org/licenses/MIT
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
"""
import logging
from dataclasses import dataclass, field
from typing import Dict, List, Type

from .exceptions import NoFatalIssueError

logger = logging.getLogger(__name__)


@dataclass
class Issue:
    # if a issue is fatal means the main function of system is not working
    fatal: bool
    description: str

    def __str__(self):
        return f"fatal<{self.fatal}>-{self.description}"


@dataclass
class DiagnosisReport:
    system_name: str
    # when a system is healthy means all the function of it could work
    # well as expected and without any issue.
    healthy: bool = True
    # alive is a key field for indicating system health
    alive: bool = True
    is_core: bool = True
    issues: List[Issue] = field(default_factory=lambda: [])

    def __str__(self):
        return f"{self.system_name}-healthy<{self.healthy}>-alive{self.alive}"

    def __post_init__(self):
        if self.issues:
            self.set_issues(self.issues)

    def set_issues(self, issues: List[Issue]):
        self.issues = issues
        self.healthy = not bool(issues)
        try:
            self.get_fatal_issues()
        except NoFatalIssueError:
            self.alive = True
        else:
            self.alive = False

    def get_fatal_issues(self) -> List[Issue]:
        fatal_issues = []
        for issue in self.issues:
            if issue.fatal:
                fatal_issues.append(issue)

        if not fatal_issues:
            raise NoFatalIssueError()
        return fatal_issues

    def get_report_message(self) -> str:
        fatal_issues = self.get_fatal_issues()
        reasons = [x.description for x in fatal_issues]
        return "; ".join(reasons)


@dataclass
class DiagnosisReportList:
    items: List[DiagnosisReport]

    @property
    def is_death(self) -> bool:
        """if any core probe give not alive diagnosis"""
        for sub_report in self.items:
            if sub_report.is_core and not sub_report.alive:
                return True

        return False

    def get_fatal_report(self) -> Dict[str, str]:
        if not self.is_death:
            raise RuntimeWarning("there is not death report")

        return {x.system_name: x.get_report_message() for x in self.items if not x.alive}


class VirtualProbe:
    """virtual probe"""

    name: str = ''
    is_core: bool = True

    def report(self) -> DiagnosisReport:
        """report diagnose report"""
        report = DiagnosisReport(system_name=self.name, is_core=self.is_core)
        issues = self.diagnose()
        if issues:
            report.set_issues(issues)
        return report

    def diagnose(self) -> List[Issue]:
        """Make a diagnosis, return the issue list."""
        raise NotImplementedError


@dataclass
class ProbeSet:
    # a ProbeSet has to own one probe at least
    probes: List[Type[VirtualProbe]]

    def examination(self) -> DiagnosisReportList:
        diagnosis_list = []
        for probe_cls in self.probes:
            diagnosis_list.append(probe_cls().report())
        return DiagnosisReportList(diagnosis_list)
